import os
import json

import dgl
import networkx as nx
from collections import defaultdict

from . import flake
from . import config

#######################
#        helpers      #
#######################
def one_hot_encode(idx, len):
    ret = [0] * len
    ret[idx] = 1
    return ret

#######################
#    output funcs     #
#######################

def print_flake_info(G, id):
    print("Statistics for graph: " + str(id))
    print("Actions taken: " + ",".join(G.getActions()))
    print("number of nodes: " + str(G.num_nodes()))
    print("number of node types: " + str(G.num_node_types()))
    for type in G.nodetypes:
        print(type + " - " + str(len(G.nodetypes[type])))
    print("number of edges: " + str(G.num_edges()))
    print("number of edge types: " + str(G.num_edge_types()))
    for type in G.edgetypes:
        print(type + " - " + str(len(G.edgetypes[type])))
    print("number of distinct edge schemas: " + str(G.num_schemas()))
    for schema in G.get_schemas():
        print(schema)

# make a graph with whatever has been loaded into our storage.
def flake_to_networkx_graph_with_attributes(G):
    print("Making NetworkX Graph...")
    nxG = nx.MultiDiGraph()
    nxG.add_nodes_from(G.get_nodes_with_attributes())
    nxG.add_edges_from(G.get_edges_with_attributes())
    print("Graph constructed.")
    return nxG

def flake_to_networkx_graph(G):
    print("Making NetworkX Graph...")
    nxG = nx.MultiDiGraph()
    nxG.add_nodes_from(G.get_nodes_for_networkx())
    nxG.add_edges_from(G.get_edges_for_networkx())
    print("Graph constructed.")
    return nxG

def flake_to_dgl_graph(G):
    print("Making DGL Graph...")
    DG = dgl.graph(G.get_graph_dictionary(), idtype=int)
    print("Graph constructed.")
    return DG

def flake_to_labeled_networkx_graph(G):
    print("Making NetworkX Graph...")
    nxG = nx.DiGraph()
    nxG.add_nodes_from(G.get_nodes_for_networkx())
    nxG.add_edges_from(G.get_edges_for_networkx())
    nx.draw_networkx_labels(nxG, pos=nx.spring_layout(nxG), labels=G.get_nodes_with_types())
    nx.draw_networkx_edge_labels(nxG, pos=nx.spring_layout(nxG), edge_labels=G.get_edges_with_types())
    print("Graph constructed.")
    return nxG

# serialize the graph with a default path name.
def flake_to_pickle(G):
    print("converting graph to pickle...")
    OUTPUT_DIR = str(os.getcwd()) + "/" + config.initFromConfig('OUTPUT_DIR')
    out = OUTPUT_DIR + "/graph{}/graph{}.gpickle".format(str(G.getIndex()), str(G.getIndex()))
    NG = flake_to_networkx_graph(G)
    nx.write_gpickle(NG, out)
    print("Graph pickle outputted to " + out + ".")

def flake_to_png(G):
    NG = nx.DiGraph()
    NG.add_nodes_from(G.get_nodes_for_networkx())
    reNG = nx.relabel_nodes(NG, G.get_nodes_with_types())
    for edge in G.get_edges():
        reNG.add_edge(edge.getSrcNode().getType(), edge.getDstNode().getType(), label=edge.getType(), data=edge.getType())
    print("Drawing graph...")
    OUTPUT_DIR = str(os.getcwd()) + "/" + config.initFromConfig('OUTPUT_DIR')
    out = OUTPUT_DIR + "/graph{}".format(str(G.getIndex()))
    if not os.path.exists(out):
        os.makedirs(out)
    out += "/graph{}.png".format(str(G.getIndex()))

    A = nx.nx_agraph.to_agraph(reNG)
    A.draw(out, format="png", prog="dot", args="-Goverlap=scale -Efont_size=8")
    print("Graph drawn to " + out + ".")

#output a JSON-formatted dictionary where key = node type, value = list of nodes
def encoded_node_types(G):
    print("outputting node types to JSON format...")
    type_dict = {}
    types = list(G.nodetypes.keys())
    for node in G.nodes:
        type_dict[G.nodes[node].id()] = one_hot_encode(types.index(G.nodes[node].getType()), len(types))
    OUTPUT_DIR = str(os.getcwd()) + "/" + config.initFromConfig('OUTPUT_DIR')
    out = OUTPUT_DIR + "/graph{}".format(str(G.getIndex()))
    if not os.path.exists(out):
        os.makedirs(out)
    out += "/nodetypes{}.json".format(str(G.getIndex()), str(G.getIndex()))
    with open(f'{out}', 'w') as f:
        f.write(str(types))
        json.dump(type_dict, f, indent=2)
    print("Node types outputted to " + out + ".")


#output a JSON-formatted dictionary where key = edge type, value = list of nodes
def encoded_edge_types(G):
    print("outputting edge types to JSON format...")
    type_dict = {}
    types = list(G.edgetypes.keys())
    for edge in G.edges:
        type_dict[G.edges[edge].id()] = one_hot_encode(types.index(G.edges[edge].getType()), len(types))

    OUTPUT_DIR = str(os.getcwd()) + "/" + config.initFromConfig('OUTPUT_DIR')
    out = OUTPUT_DIR + "/graph{}".format(str(G.getIndex()))
    if not os.path.exists(out):
        os.makedirs(out)
    out += "/edgetypes{}.json".format(str(G.getIndex()), str(G.getIndex()))
    with open(f'{out}', 'w') as f:
        f.write(str(types))
        json.dump(type_dict, f, indent=2)
    print("Edge types outputted to " + out + ".")

#return a python dictionary of the graph
def flake_to_dictionary(G):
    return G.get_graph_dictionary()

# output a JSON-formatted dictionary where key = edge_schema, value = src node list and dest node list where src[i]-edge-dest[i]
def flake_to_json(G):
    print("outputting graph to JSON format...")
    output_dict = G.get_graph_dictionary()
    OUTPUT_DIR = str(os.getcwd()) + "/" + config.initFromConfig('OUTPUT_DIR')
    out = OUTPUT_DIR + "/graph{}".format(str(G.getIndex()))
    if not os.path.exists(out):
        os.makedirs(out)
    out += "/graph{}.json".format(str(G.getIndex()), str(G.getIndex()))
    with open(f'{out}', 'w') as f:
        json.dump(output_dict, f)
    print("Graph outputted to " + out + ".")

# print graph info to a file
def flake_info_to_file(G):
    OUTPUT_DIR = str(os.getcwd()) + "/" + config.initFromConfig('OUTPUT_DIR')
    out = OUTPUT_DIR + "/graph{}".format(str(G.getIndex()))
    if not os.path.exists(out):
        os.makedirs(out)
    out += "/stats{}.txt".format(str(G.getIndex()), str(G.getIndex()))
    with open(f'{out}', 'w') as f:
        f.write("Statistics for graph: " + str(G.getIndex()) + "\n")
        f.write("Actions taken: " + ",".join(G.getActions()))
        f.write("number of nodes: " + str(G.num_nodes()) + "\n")
        f.write("number of node types: " + str(G.num_node_types()) + "\n")
        for type in G.nodetypes:
            f.write(type + " - " + str(len(G.nodetypes[type])) + "\n")
        f.write("number of edges: " + str(G.num_edges()) + "\n")
        f.write("number of edge types: " + str(G.num_edge_types()) + "\n")
        for type in G.edgetypes:
            f.write(type + " - " + str(len(G.edgetypes[type])) + "\n")
        f.write("number of distinct edge schemas: " + str(G.num_schemas()) + "\n")
        for schema in G.get_schemas():
            str_schema = '-'.join(schema)
            f.write(str(str_schema) + " - " + str(len(G.schemas[schema])) + "\n")



import paho.mqtt.client as mqtt
import sys
import subprocess
import base64
import zlib
import sqlite3
import time
from termcolor import colored

from . import config
from . import queries
from . import flake

class FlurryBank():
    def __init__(self, filter):
        DB_FILE = str(os.getcwd() + "/" + str(config.initFromConfig('DB_FILE')))
        if DB_FILE is not None:
            self.db_file = DB_FILE
            self.client = mqtt.Client()
            db_conn = sqlite3.connect(self.db_file)
            cursor = db_conn.cursor()
            cursor.execute(queries.create_type_table())
            cursor.execute(queries.create_node_table())
            cursor.execute(queries.create_edge_table())
            cursor.execute(queries.create_graph_table())
            db_conn.commit()
            cursor.close()
        else:
            print("Could not find database file from configuration.")
            sys.exit()
        self.filter = filter

    def make_flake(self, actions):
        db_conn = sqlite3.connect(self.db_file)
        cursor = db_conn.cursor()
        action_str = ",".join(actions)
        sql = queries.insert_graph(action_str)
        cursor.execute(sql)
        db_conn.commit()
        cursor.execute(queries.get_last_row_id())
        id = cursor.fetchall()[0][0]
        cursor.close()
        return flake.FlurryFlake(id, actions)

    #def get_flake(self, id):


    def on_connect(self, client, userdata, flags, rc):
        print("Connected with result code "+str(rc))
        topic = config.initFromConfig('MQTT_TOPIC')
        if topic is not None:
            client.subscribe(topic, qos=0)
        else:
            print("MQTT topic improperly configured, exiting.")
            sys.exit()

    def on_message(self, client, userdata, msg):
        decoded_msg = zlib.decompress(base64.b64decode(msg.payload.decode('latin-1'))).decode('latin-1')
        self.filter.load_data(decoded_msg, userdata['graph'])

    def on_disconnect(self, client, userdata, rc=0):
        print("disconnected with result code "+ str(rc))
        self.client.loop_stop()

    def connect_client(self, graph):
        print("Connecting MQTT subscriber...")
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.on_disconnect = self.on_disconnect

        user = config.initFromConfig('MQTT_USERNAME')
        passwd = config.initFromConfig('MQTT_PASSWORD')
        host = config.initFromConfig('MQTT_HOST')
        port = config.initFromConfig('MQTT_PORT')

        self.client.username_pw_set(user, passwd)
        self.client.connect(host, int(port), 60)
        self.client.user_data_set({'graph': graph})
        self.client.loop_start()
        time.sleep(1)

    def disconnect_client(self, graph):
        print("Stopping MQTT subscriber...")
        self.client.loop_stop()
        db_conn = sqlite3.connect(self.db_file)
        save = config.initFromConfig("SAVE_TO_DISK")
        if save == 'yes':
            graph.save_to_disk(db_conn)
        graph.to_png()
        graph.to_pickle()
        graph.to_file()
        graph.to_edge_type_dictionary()
        graph.to_node_type_dictionary()
        graph.to_json()



import orjson
import os
import sys
import time

from collections import defaultdict
from flake.filters.filter import Filter

master_node_dict = defaultdict()

class W3CFilter(Filter):
    """
    Initialize the filter.
    """
    def __init__(self):
        Filter.__init__(self)

    def parse_nodes_for_edge(self, prov_type, values):
        src_node = ""
        dst_node = ""

        if prov_type == "wasGeneratedBy":
            src_node = values["prov:entity"]
            dst_node = values["prov:activity"]
        elif prov_type == "used":
            src_node = values["prov:activity"]
            dst_node = values["prov:entity"]
        elif prov_type == "wasInformedBy":
            src_node = values["prov:informant"]
            dst_node = values["prov:informed"]
        elif prov_type == "wasInfluencedBy":
            src_node = values["prov:influencee"]
            dst_node = values["prov:influencer"]
        elif prov_type == "wasAssociatedWith":
            src_node = values["prov:agent"]
            dst_node = values["prov:activity"]
        elif prov_type == "wasDerivedFrom":
            src_node = values["prov:generatedEntity"]
            dst_node = values["prov:usedEntity"]
        # not currently supported by CamFlow
        elif prov_type == "wasStartedBy" or prov_type == "wasEndedBy":
            src_node = values["prov:trigger"]
            dst_node = values["prov:activity"]
        elif prov_type == "wasInvalidatedBy":
            src_node = values["prov:activity"]
            dst_node = values["prov:entity"]
        elif prov_type == "wasAttributedTo":
            src_node = values["prov:entity"]
            dst_node = values["prov:agent"]
        elif prov_type == "actedOnBehalfOf":
            src_node = values["prov:delegate"]
            dst_node = values["prov:responsible"]
        elif prov_type == "specializationOf":
            src_node = values["prov:specificEntity"]
            dst_node = values["prov:generalEntity"]
        return src_node, dst_node

    def load_data(self, data, G):
        object = orjson.loads(data)
        features = {}
        type = ""

        for prov_type in object:

            if prov_type == "activity":
                for node in object[prov_type]:
                    features = object[prov_type][node]
                    if self.node_granularity == "fine":
                        type = object[prov_type][node]["prov:type"]
                    else:
                        type = prov_type
                    id = G.add_node(type, features)
                    master_node_dict[node] = id

            elif prov_type == "entity":
                for node in object[prov_type]:
                    if "cf:camflow" in object[prov_type][node]:
                        continue
                    features = object[prov_type][node]
                    if self.node_granularity == "fine":
                        type = object[prov_type][node]["prov:type"]
                    else:
                        type = prov_type
                    id = G.add_node(type, features)
                    master_node_dict[node] = id

            elif prov_type == "agent":
                for node in object[prov_type]:
                    features = object[prov_type][node]
                    if self.node_granularity == "fine":
                        type = object[prov_type][node]["prov:type"]
                    else:
                        type = prov_type
                    id = G.add_node(type, features)
                    master_node_dict[node] = id

            elif prov_type == "prefix":
                continue
            else:
                for edge in object[prov_type]:
                    features = object[prov_type][edge]
                    jiffies = int(object[prov_type][edge]["cf:jiffies"])
                    if self.edge_granularity == "fine":
                        type = object[prov_type][edge]["prov:type"]
                    else:
                        type = prov_type
                    src_node, dst_node = self.parse_nodes_for_edge(prov_type, features)
                    if src_node in master_node_dict and dst_node in master_node_dict:
                        G.add_edge(type, master_node_dict[src_node], master_node_dict[dst_node], features, jiffies)

    """
    Load the specified graph from file, using the filter to parse the data.
    params: input_path, the full file path to the camflow data.
    params: G, the flake object we are loading to
    """
    def load_data_from_file(self, data, G):
        print("Loading data now...")
        logcount = 0
        start = time.time()
        for line in data:
            load_data(line[0], G)
            logcount +=1
            if logcount % 5000 == 0:
                end = time.time()
                hours, rem = divmod(end-start, 3600)
                minutes, seconds = divmod(rem, 60)
                print("loaded " + str(round(((logcount/len(json_data))*100), 2)) + "% of data in " +
                    ("{:0>2}:{:0>2}:{:05.2f}".format(int(hours),int(minutes),seconds)) + " seconds.")
                start = time.time()
        print("Data loaded! Added " + str(logcount) + " logs.")



import time
import os
import json
import dgl
import networkx as nx
from collections import defaultdict

from . import config as cfg
from . import queries
from . import node
from . import edge

"""
Helper function to return a one-hot encoded vector
"""
def one_hot_encode(idx, len):
    ret = [0] * len
    ret[idx] = 1
    return ret

"""
This class represents an internal graph of Nodes and Edges
"""
class FlurryFlake():
    def __init__(self, id, actions):
        # graph id
        self.id = id
        # descriptive list of the actions performed
        self.actions = actions
        # node serial to Node object
        self.nodes = defaultdict()
        # edge serial to Edge object
        self.edges = defaultdict()
        # type schema (nt-et-nt), to list of node index tuples (src,dst)
        self.schemas = defaultdict(list)
        self.nodetypes = defaultdict(list)
        self.edgetypes = defaultdict(list)

    def getIndex(self):
        return self.id

    def getActions(self):
        return self.actions

    def num_nodes(self):
        return len(self.nodes)

    def num_edges(self):
        return len(self.edges)

    def num_node_types(self):
        return len(self.nodetypes)

    def num_edge_types(self):
        return len(self.edgetypes)

    def num_schemas(self):
        return len(self.schemas)

    def num_nodes_of_type(self, type):
        if type in self.nodetypes:
            return len(self.nodetypes[type])
        return 0

    def num_edges_of_type(self, type):
        if type in self.edgetypes:
            return len(self.edgetypes[type])
        return 0

    def num_edges_of_schema(self, schema):
        if schema in self.schemas:
            return len(self.schemas[schema])
        return 0

    #returns list of node ids
    def get_nodes_for_networkx(self):
        node_list = []
        for node in self.nodes:
            node_list.append(int(self.nodes[node].id()))
        return node_list

    def get_node_types(self):
        return self.nodetypes

    def get_nodes_with_types(self):
        node_dict = {}
        for node in self.nodes:
            node_dict[int(self.nodes[node].id())] = self.nodes[node].getType()
        return node_dict

    def get_nodes_encoded(self):
        type_dict = {}
        types = list(self.nodetypes.keys())
        for node in self.nodes:
            type_dict[self.nodes[node].id()] = one_hot_encode(types.index(self.nodes[node].getType()), len(types))
        return type_dict

    def get_nodes_with_attributes(self):
        node_list = []
        for node in self.nodes:
            node_list.append((self.nodes[node].id(), self.nodes[node].getFeatures()))
        return node_list

    def get_edges(self):
        edge_list = []
        for edge in self.edges:
            edge_list.append(self.edges[edge])
        return edge_list

    def get_edges_for_networkx(self):
        edge_list = []
        for edge in self.edges:
            edge_list.append((int(self.edges[edge].getSrcNode().id()),
                              int(self.edges[edge].getDstNode().id())))
        return edge_list

    def get_edges_with_types(self):
        edge_dict = {}
        for edge in self.edges:
            edge_dict[(int(self.edges[edge].getSrcNode().id()),
                       int(self.edges[edge].getDstNode().id()))] = self.edges[edge].getType()
        return edge_dict

    def get_edges_encoded(self):
        type_dict = {}
        types = list(self.edgetypes.keys())
        for edge in self.edges:
            type_dict[self.edges[edge].id()] = one_hot_encode(types.index(self.edges[edge].getType()), len(types))
        return type_dict

    def get_edges_with_attributes(self):
        edge_list = []
        for edge in self.edges:
            edge_list.append((int(self.edges[edge].getSrcNode().id()),
                              int(self.edges[edge].getDstNode().id()),
                              self.edges[edge].getFeatures()))
        return edge_list

    def get_schemas(self):
        return list(self.schemas.keys())

    def get_schema_node_lists(self, schema):
        return list(map(list, zip(*self.schemas[schema])))

    def get_graph_dictionary(self):
        output_dict = defaultdict(list)
        for schema in self.get_schemas():
            str_schema = '-'.join(schema)
            src_dst = self.get_schema_node_lists(schema)
            output_dict[str_schema] = (src_dst[0], src_dst[1])
        return output_dict

    def add_node(self, type):
        id = len(self.nodes)
        self.nodes[id] = node.Node(type, id)
        self.nodetypes[type].append(id)
        return id

    def add_node(self, type, features):
        id = len(self.nodes)
        self.nodes[id] = node.Node(type, id)
        self.nodetypes[type].append(id)
        self.nodes[id].add_features(features)
        return id

    def add_edge(self, type, src_node, dst_node, jiffies):
        id = len(self.edges)
        self.edges[id] = edge.Edge(type, self.nodes[src_node], self.nodes[dst_node], jiffies, id)
        self.edgetypes[type].append(id)
        return id

    def add_edge(self, type, src_node, dst_node, features, jiffies):
        id = len(self.edges)
        self.edges[id] = edge.Edge(type, self.nodes[src_node], self.nodes[dst_node], jiffies, id)
        self.edgetypes[type].append(id)
        self.edges[id].add_features(features)
        # add edge schema
        self.schemas[(self.nodes[src_node].getType(),
            type,
            self.nodes[dst_node].getType())].append((self.nodes[src_node].id(),
            self.nodes[dst_node].id()))
        return id

    def save_to_disk(self, db):
        start = time.time()
        print("Saving flake to database...")
        cursor = db.cursor()
        for type in self.nodetypes.keys():
            sql = queries.insert_type(type)
            cursor.execute(sql)
        for type in self.edgetypes.keys():
            sql = queries.insert_type(type)
            cursor.execute(sql)
        db.commit()
        print("type: " + str(time.time() - start))

        start = time.time()
        edgefeatures = []
        #cursor.execute("BEGIN TRANSACTION;")
        for edge in self.edges:
            sql = queries.get_type_index(self.edges[edge].getSrcNode().getType())
            cursor.execute(sql)
            src_node_type_id = cursor.fetchone()[0]
            sql = queries.insert_node(src_node_type_id, self.getIndex())
            cursor.execute(sql)
            cursor.execute(queries.get_last_row_id())
            src_node_id = cursor.fetchone()[0]

            sql = queries.get_type_index(self.edges[edge].getDstNode().getType())
            cursor.execute(sql)
            dst_node_type_id = cursor.fetchone()[0]
            sql = queries.insert_node(dst_node_type_id, self.getIndex())
            cursor.execute(sql)
            cursor.execute(queries.get_last_row_id())
            dst_node_id = cursor.fetchone()[0]

            start = time.time()

            sql = queries.get_type_index(self.edges[edge].getType())
            cursor.execute(sql)
            edge_type_id = cursor.fetchone()[0]

            sql = queries.insert_edge(self.getIndex(),
                edge_type_id,
                src_node_id,
                dst_node_id,
                self.edges[edge].getJiffies())
            cursor.execute(sql)
        print("edges: " + str(time.time()-start))
        db.commit()
        cursor.close()
        print("Graph {} saved.".format(self.id))

    def clear(self):
        self.nodes = defaultdict()
        self.edges = defaultdict()
        self.schemas = defaultdict(list)
        self.nodetypes = defaultdict(list)
        self.edgetypes = defaultdict(list)

##########################
#    Output functions    #
##########################

    def print_info(self):
        print("Statistics for graph: " + str(self.id))
        print("Actions taken: " + str(",".join(self.getActions())))
        print("number of nodes: " + str(self.num_nodes()))
        print("number of node types: " + str(self.num_node_types()))
        for type in self.nodetypes:
            print(type + " - " + str(len(self.nodetypes[type])))
        print("number of edges: " + str(self.num_edges()))
        print("number of edge types: " + str(self.num_edge_types()))
        for type in self.edgetypes:
            print(type + " - " + str(len(self.edgetypes[type])))
        print("number of distinct edge schemas: " + str(self.num_schemas()))
        for schema in self.get_schemas():
            print(schema)

    def to_networkx_graph(self):
        print("Making NetworkX Graph...")
        nxG = nx.MultiDiGraph()
        nxG.add_nodes_from(self.get_nodes_for_networkx())
        nxG.add_edges_from(self.get_edges_for_networkx())
        print("Graph constructed.")
        return nxG

    # make a graph with whatever has been loaded into our storage.
    def to_networkx_graph_with_attributes(self):
        print("Making NetworkX Graph...")
        nxG = nx.MultiDiGraph()
        nxG.add_nodes_from(self.get_nodes_with_attributes())
        nxG.add_edges_from(self.get_edges_with_attributes())
        print("Graph constructed.")
        return nxG

    def to_networkx_graph_with_labels(self):
        print("Making NetworkX Graph...")
        nxG = nx.DiGraph()
        nxG.add_nodes_from(self.get_nodes_for_networkx())
        nxG.add_edges_from(self.get_edges_for_networkx())
        nx.draw_networkx_labels(nxG, pos=nx.spring_layout(nxG), labels=self.get_nodes_with_types())
        nx.draw_networkx_edge_labels(nxG, pos=nx.spring_layout(nxG), edge_labels=self.get_edges_with_types())
        print("Graph constructed.")
        return nxG

    def to_dgl_graph(self):
        print("Making DGL Graph...")
        DG = dgl.graph(self.get_graph_dictionary(), idtype=int)
        print("Graph constructed.")
        return DG

    def to_dictionary(self):
        return self.get_graph_dictionary()

    def to_json(self):
        print("outputting graph to JSON format...")
        output_dict = self.get_graph_dictionary()
        OUTPUT_DIR = cfg.initFromConfig('OUTPUT_DIR')
        out = OUTPUT_DIR + "/graph{}".format(str(self.getIndex()))
        if not os.path.exists(out):
            os.makedirs(out)
        out += "/graph{}.json".format(str(self.getIndex()), str(self.getIndex()))
        with open(f'{out}', 'w') as f:
            json.dump(output_dict, f)
        print("Graph outputted to " + out + ".")

    def to_file(self):
        OUTPUT_DIR = cfg.initFromConfig('OUTPUT_DIR')
        out = OUTPUT_DIR + "/graph{}".format(str(self.getIndex()))
        if not os.path.exists(out):
            os.makedirs(out)
        out += "/stats{}.txt".format(str(self.getIndex()), str(self.getIndex()))
        with open(f'{out}', 'w') as f:
            f.write("Statistics for graph: " + str(self.getIndex()) + "\n")
            f.write("Actions taken: " + ",".join(self.getActions()) + "\n")
            f.write("number of nodes: " + str(self.num_nodes()) + "\n")
            f.write("number of node types: " + str(self.num_node_types()) + "\n")
            for type in self.nodetypes:
                f.write(type + " - " + str(len(self.nodetypes[type])) + "\n")
            f.write("number of edges: " + str(self.num_edges()) + "\n")
            f.write("number of edge types: " + str(self.num_edge_types()) + "\n")
            for type in self.edgetypes:
                f.write(type + " - " + str(len(self.edgetypes[type])) + "\n")
            f.write("number of distinct edge schemas: " + str(self.num_schemas()) + "\n")
            for schema in self.get_schemas():
                str_schema = '-'.join(schema)
                f.write(str(str_schema) + " - " + str(len(self.schemas[schema])) + "\n")

    def to_pickle(self):
        print("converting graph to pickle...")
        OUTPUT_DIR = cfg.initFromConfig('OUTPUT_DIR')
        out = OUTPUT_DIR + "/graph{}".format(str(self.getIndex()))
        if not os.path.exists(out):
            os.makedirs(out)
        out += "/graph{}.gpickle".format(str(self.getIndex()), str(self.getIndex()))
        NG = self.to_networkx_graph()
        nx.write_gpickle(NG, out)
        print("Graph pickle outputted to " + out + ".")

    def to_png(self):
        NG = nx.DiGraph()
        NG.add_nodes_from(self.get_nodes_for_networkx())
        reNG = nx.relabel_nodes(NG, self.get_nodes_with_types())
        for edge in self.get_edges():
            reNG.add_edge(edge.getSrcNode().getType(), edge.getDstNode().getType(), label=edge.getType(), data=edge.getType())
        print("Drawing graph...")
        OUTPUT_DIR = cfg.initFromConfig('OUTPUT_DIR')
        out = OUTPUT_DIR + "/graph{}".format(str(self.getIndex()))
        if not os.path.exists(out):
            os.makedirs(out)
        out += "/graph{}.png".format(str(self.getIndex()))

        A = nx.nx_agraph.to_agraph(reNG)
        A.draw(out, format="png", prog="dot", args="-Goverlap=scale -Efont_size=8")
        print("Graph drawn to " + out + ".")

    def to_edge_type_dictionary(self):
        print("outputting one-hot encoded edge types to JSON format...")
        types = list(self.edgetypes.keys())
        OUTPUT_DIR = cfg.initFromConfig('OUTPUT_DIR')
        out = OUTPUT_DIR + "/graph{}".format(str(self.getIndex()))
        if not os.path.exists(out):
            os.makedirs(out)
        out += "/edgetypes{}.json".format(str(self.getIndex()), str(self.getIndex()))
        with open(f'{out}', 'w') as f:
            f.write(str(types))
            json.dump(self.get_edges_encoded(), f, indent=2)
        print("Edge types outputted to " + out + ".")

    def to_node_type_dictionary(self):
        print("outputting one-hot encoded node types to JSON format...")
        types = list(self.nodetypes.keys())
        OUTPUT_DIR = cfg.initFromConfig('OUTPUT_DIR')
        out = OUTPUT_DIR + "/graph{}".format(str(self.getIndex()))
        if not os.path.exists(out):
            os.makedirs(out)
        out += "/nodetypes{}.json".format(str(self.getIndex()), str(self.getIndex()))
        with open(f'{out}', 'w') as f:
            f.write(str(types))
            json.dump(self.get_nodes_encoded(), f, indent=2)
        print("Node types outputted to " + out + ".")
